import 'package:firebase_auth/firebase_auth.dart';
import 'package:firebase_auth_helper/firebase_auth_helper.dart';
import 'package:firebase_auth_helper/src/user_meta_data.dart';
import 'package:flutter_test/flutter_test.dart';
import 'package:mockito/annotations.dart';
import 'package:mockito/mockito.dart';

import 'firebase_auth_provider_test.mocks.dart';

const uid = "user123";
const displayName = "John";
const email = "john.potter@gmail.com";
const emailVerified = true;
const isAnonymous = false;
final creationTime = DateTime(2002, 4, 4, 1, 1);
final lastSignInTime = DateTime(2003, 5, 5, 1, 1);
final metaData = UserMetaData(creationTime, lastSignInTime);
final metadata = MockUserMetadata();
const phoneNumber = "+412223344";
const photoUrl = "someUrl";
const providerData = <UserInfo>[];
const refreshToken = "someRefreshToken";

@GenerateMocks([FirebaseAuth, User, UserMetadata])
main() {
  late MockFirebaseAuth firebaseAuth;
  late MockUser loggedInUser;

  late FirebaseAuthProvider authProvider;

  setUp(() {
    firebaseAuth = MockFirebaseAuth();
    authProvider = FirebaseAuthProvider(firebaseAuth);

    loggedInUser = MockUser();

    when(metadata.creationTime).thenReturn(creationTime);
    when(metadata.lastSignInTime).thenReturn(lastSignInTime);

    when(loggedInUser.uid).thenReturn(uid);
    when(loggedInUser.displayName).thenReturn(displayName);
    when(loggedInUser.email).thenReturn(email);
    when(loggedInUser.emailVerified).thenReturn(emailVerified);
    when(loggedInUser.isAnonymous).thenReturn(isAnonymous);
    when(loggedInUser.metadata).thenReturn(metadata);
    when(loggedInUser.phoneNumber).thenReturn(phoneNumber);
    when(loggedInUser.photoURL).thenReturn(photoUrl);
    when(loggedInUser.providerData).thenReturn(providerData);
    when(loggedInUser.refreshToken).thenReturn(refreshToken);
    when(loggedInUser.updateDisplayName(any))
        .thenAnswer((_) => Future.delayed(Duration.zero));
    when(loggedInUser.sendEmailVerification(any))
        .thenAnswer((_) => Future.delayed(Duration.zero));
    when(loggedInUser.reload())
        .thenAnswer((_) => Future.delayed(Duration.zero));
    when(loggedInUser.getIdToken(false))
        .thenAnswer((_) => Future.value("some token"));
  });

  AuthUser mappedUser(User user) {
    return AuthUser(user.uid,
        displayName: user.displayName ?? "",
        email: user.email ?? "",
        emailVerified: user.emailVerified,
        isAnonymous: user.isAnonymous,
        metaData: UserMetaData.fromUserMetadata(user.metadata),
        phoneNumber: user.phoneNumber,
        photoURL: user.photoURL,
        providerData: user.providerData,
        refreshToken: user.refreshToken,
        updateDisplayName: user.updateDisplayName,
        sendVerificationEmail: user.sendEmailVerification,
        reload: user.reload,
        getIdToken: user.getIdToken);
  }

  TypeMatcher<AuthUser> authUserMatcher(User original) => isA<AuthUser>()
      .having((p0) => p0.id, "id", original.uid)
      .having((p0) => p0.displayName, "displayName", original.displayName)
      .having((p0) => p0.email, "email", original.email)
      .having((p0) => p0.emailVerified, "emailVerified", original.emailVerified)
      .having((p0) => p0.isAnonymous, "isAnonymous", original.isAnonymous)
      .having(
          (p0) => p0.metaData,
          "metadata",
          isA<UserMetaData>()
              .having((p0) => p0.creationTime, "creationTime", creationTime)
              .having(
                  (p0) => p0.lastSignInTime, "lastSignInTime", lastSignInTime))
      .having((p0) => p0.phoneNumber, "phoneNumber", original.phoneNumber)
      .having((p0) => p0.photoURL, "photoURL", original.photoURL)
      .having((p0) => p0.providerData, "providerData", original.providerData)
      .having((p0) => p0.refreshToken, "refreshToken", original.refreshToken);

  void expectAuthUserResult(AuthUser actual, MockUser loggedInUser) {
    final expected = mappedUser(loggedInUser);

    expect(expected, authUserMatcher(loggedInUser));
    actual.updateDisplayName("New John");
    actual.sendVerificationEmail();
    actual.reload();
    actual.getIdToken();

    verifyInOrder([
      loggedInUser.updateDisplayName("New John"),
      loggedInUser.sendEmailVerification(),
      loggedInUser.reload(),
      loggedInUser.getIdToken(),
    ]);
  }

  test("Get current auth user forwards to firebaseAuth and maps correctly", () {
    when(firebaseAuth.currentUser).thenReturn(loggedInUser);
    expectAuthUserResult(authProvider.currentAuthUser!, loggedInUser);
  });

  test("Get current auth user returns null if firebaseAuth returns null", () {
    when(firebaseAuth.currentUser).thenReturn(null);
    expect(authProvider.currentAuthUser, null);
  });

  test("authStateChanged emits events when firebaseAuth state changed",
      () async {
    when(firebaseAuth.authStateChanges())
        .thenAnswer((_) => Stream.fromIterable([
              null,
              loggedInUser,
              null,
            ]));
    final result = authProvider.authStateChanged;

    expect(result, emitsInOrder([null, authUserMatcher(loggedInUser), null]));
  });

  test("userChanges emits events when firebaseAuth user changes", () async {
    when(firebaseAuth.userChanges()).thenAnswer((_) => Stream.fromIterable([
          null,
          loggedInUser,
          null,
        ]));
    final result = authProvider.userChanges;

    expect(result, emitsInOrder([null, authUserMatcher(loggedInUser), null]));
  });
}
